import time
import asyncpg

from bbot.modules.templates.subdomain_enum import subdomain_enum


class crt_db(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Query crt.sh (certificate transparency) for subdomains via PostgreSQL",
        "created_date": "2025-03-27",
        "author": "@TheTechromancer",
    }

    deps_pip = ["asyncpg"]

    db_host = "crt.sh"
    db_port = 5432
    db_user = "guest"
    db_name = "certwatch"
    reject_wildcards = False

    async def setup(self):
        self.db_conn = None
        return await super().setup()

    async def request_url(self, query):
        if not self.db_conn:
            self.db_conn = await asyncpg.connect(
                host=self.db_host,
                port=self.db_port,
                user=self.db_user,
                database=self.db_name,
                statement_cache_size=0,  # Disable automatic statement preparation
            )

        sql = """
        WITH ci AS (
            SELECT array_agg(DISTINCT sub.NAME_VALUE) NAME_VALUES
            FROM (
                SELECT DISTINCT cai.CERTIFICATE, cai.NAME_VALUE
                FROM certificate_and_identities cai
                WHERE plainto_tsquery('certwatch', $1) @@ identities(cai.CERTIFICATE)
                    AND cai.NAME_VALUE ILIKE ('%.' || $1)
                LIMIT 50000
            ) sub
            GROUP BY sub.CERTIFICATE
        )
        SELECT DISTINCT unnest(NAME_VALUES) as name_value FROM ci;
        """
        start = time.time()
        results = await self.db_conn.fetch(sql, query)
        end = time.time()
        self.verbose(f"SQL query executed in: {end - start} seconds with {len(results):,} results")
        return results

    async def parse_results(self, results, query):
        domains = set()
        for row in results:
            domain = row["name_value"]
            if domain:
                for d in domain.splitlines():
                    domains.add(d.lower())
        return domains

    async def cleanup(self):
        if self.db_conn:
            await self.db_conn.close()



from .base import ModuleTestBase

# import logging


class TestNewsletters(ModuleTestBase):
    found_tgt = "http://127.0.0.1:8888/found"
    missing_tgt = "http://127.0.0.1:8888/missing"
    targets = [found_tgt, missing_tgt]
    modules_overrides = ["speculate", "httpx", "newsletters"]

    html_with_newsletter = """
    <input aria-required="true"
    class="form-input form-input-text required"
    data-at="form-email"
    data-describedby="form-validation-error-box-element-5"
    data-label-inside="Enter your email"
    id="field-5f329905b4bfe1027b44513f94b50363-0"
    name="Enter your email"
    placeholder="Enter your email"
    required=""
    title="Enter your email"
    type="email" value=""/>
    """

    html_without_newsletter = """
    <div>
    <h1>Example Domain</h1>
    <p>This domain is for use in illustrative examples in documents. You may use this
    domain in literature without prior coordination or asking for permission.</p>
    <p><a href="https://www.iana.org/domains/example">More information...</a></p>
    </div>
    """

    async def setup_after_prep(self, module_test):
        request_args = {"uri": "/found", "headers": {"test": "header"}}
        respond_args = {"response_data": self.html_with_newsletter}
        module_test.set_expect_requests(request_args, respond_args)
        request_args = {"uri": "/missing", "headers": {"test": "header"}}
        respond_args = {"response_data": self.html_without_newsletter}
        module_test.set_expect_requests(request_args, respond_args)

    def check(self, module_test, events):
        found = False
        missing = True
        for event in events:
            # self.log.info(f"event type: {event.type}")
            if event.type == "FINDING":
                # self.log.info(f"event data: {event.data}")
                # Verify Positive Result
                if event.data["url"] == self.found_tgt:
                    found = True
                # Verify Negative Result (should skip this statement if correct)
                elif event.data["url"] == self.missing_tgt:
                    missing = False
        assert found, "NEWSLETTER 'Found' Error - Expect status of True but got False"
        assert missing, "NEWSLETTER 'Missing' Error - Expect status of True but got False"



import logging
import traceback
import threading
from multiprocessing.context import SpawnProcess

from .misc import in_exception_chain


class BBOTThread(threading.Thread):
    default_name = "default bbot thread"

    def __init__(self, *args, **kwargs):
        self.custom_name = kwargs.pop("custom_name", self.default_name)
        if "daemon" not in kwargs:
            kwargs["daemon"] = True
        super().__init__(*args, **kwargs)

    def run(self):
        from setproctitle import setthreadtitle

        setthreadtitle(str(self.custom_name))
        super().run()


class BBOTProcess(SpawnProcess):
    default_name = "bbot process pool"

    def __init__(self, *args, **kwargs):
        self.log_queue = kwargs.pop("log_queue", None)
        self.log_level = kwargs.pop("log_level", None)
        self.custom_name = kwargs.pop("custom_name", self.default_name)
        super().__init__(*args, **kwargs)
        self.daemon = True

    def run(self):
        """
        A version of Process.run() with BBOT logging and better error handling
        """
        log = logging.getLogger("bbot.core.process")
        try:
            if self.log_level is not None and self.log_queue is not None:
                from bbot.core import CORE

                CORE.logger.setup_queue_handler(self.log_queue, self.log_level)
            if self.custom_name:
                from setproctitle import setproctitle

                setproctitle(str(self.custom_name))
            super().run()
        except BaseException as e:
            if not in_exception_chain(e, (KeyboardInterrupt,)):
                log.warning(f"Error in {self.name}: {e}")
            log.trace(traceback.format_exc())



from .base import ModuleTestBase
from bbot.test.bbot_fixtures import bbot_test_dir


class TestBucket_File_Enum(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["bucket_file_enum", "filedownload", "httpx", "excavate", "cloudcheck"]

    download_dir = bbot_test_dir / "test_bucket_file_enum"
    config_overrides = {
        "scope": {"report_distance": 5},
        "modules": {"filedownload": {"output_folder": str(download_dir)}},
    }

    open_bucket_url = "https://testbucket.s3.amazonaws.com/"
    open_bucket_body = """<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><Name>testbucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><IsTruncated>false</IsTruncated><Contents><Key>index.html</Key><LastModified>2023-05-22T23:04:38.000Z</LastModified><ETag>&quot;4a2d2d114f3abf90f8bd127c1f25095a&quot;</ETag><Size>5</Size><StorageClass>STANDARD</StorageClass></Contents><Contents><Key>test.pdf</Key><LastModified>2022-04-30T21:13:40.000Z</LastModified><ETag>&quot;723b0018c2f5a7ef06a34f84f6fa97e4&quot;</ETag><Size>388901</Size><StorageClass>STANDARD</StorageClass></Contents></ListBucketResult>"""

    pdf_data = """%PDF-1.
1 0 obj<</Pages 2 0 R>>endobj
2 0 obj<</Kids[3 0 R]/Count 1>>endobj
3 0 obj<</Parent 2 0 R>>endobj
trailer <</Root 1 0 R>>"""

    async def setup_before_prep(self, module_test):
        module_test.httpserver.expect_request("/").respond_with_data(f'<a href="{self.open_bucket_url}"/>')
        module_test.httpx_mock.add_response(
            url=self.open_bucket_url,
            text=self.open_bucket_body,
        )
        module_test.httpx_mock.add_response(
            url=f"{self.open_bucket_url}test.pdf",
            text=self.pdf_data,
            headers={"Content-Type": "application/pdf"},
        )
        module_test.httpx_mock.add_response(
            url=f"{self.open_bucket_url}test.css",
            text="",
        )

    def check(self, module_test, events):
        files = list((self.download_dir / "filedownload").glob("*.pdf"))
        assert any(e.type == "URL_UNVERIFIED" and e.data.endswith("test.pdf") for e in events)
        assert not any(e.type == "URL_UNVERIFIED" and e.data.endswith("test.css") for e in events)
        assert any(f.name.endswith("test.pdf") for f in files), "Failed to download PDF file from open bucket"
        assert not any(f.name.endswith("test.css") for f in files), "Unwanted CSS file was downloaded"



from .base import ModuleTestBase, tempwordlist


class TestBadSecrets(ModuleTestBase):
    targets = [
        "http://127.0.0.1:8888/",
        "http://127.0.0.1:8888/test.aspx",
        "http://127.0.0.1:8888/cookie.aspx",
        "http://127.0.0.1:8888/cookie2.aspx",
        "http://127.0.0.1:8888/cookie3.aspx",
    ]

    sample_viewstate = """
    <form method="post" action="./query.aspx" id="form1">
<div class="aspNetHidden">
<input type="hidden" name="__VIEWSTATE" id="__VIEWSTATE" value="rJdyYspajyiWEjvZ/SMXsU/1Q6Dp1XZ/19fZCABpGqWu+s7F1F/JT1s9mP9ED44fMkninhDc8eIq7IzSllZeJ9JVUME41i8ozheGunVSaESf4nBu" />
</div>

<div class="aspNetHidden">

    <input type="hidden" name="__VIEWSTATEGENERATOR" id="__VIEWSTATEGENERATOR" value="EDD8C9AE" />
    <input type="hidden" name="__VIEWSTATEENCRYPTED" id="__VIEWSTATEENCRYPTED" value="" />
</div>
    </form>
</body>
</html>
"""

    sample_jsf_notvuln = """
<p><input type="hidden" name="javax.faces.ViewState" id="j_id__v_0:javax.faces.ViewState:1" value="AHo0wmLu5ceItIi+I7XkEi1GAb4h12WZ894pA+Z4OH7bco2jXEy1RSCWwjtJcZNbWPcvPqL5zzfl03DoeMZfGGX7a9PSv+fUT8MAeKNouAGj1dZuO8srXt8xZIGg+wPCWWCzcX6IhWOtgWUwiXeSojCDTKXklsYt+kAAAAk5wOsXvb2lTJoO0Q==" autocomplete="off" />
"""

    modules_overrides = ["badsecrets", "httpx"]

    async def setup_after_prep(self, module_test):
        expect_args = {"uri": "/test.aspx"}
        respond_args = {"response_data": self.sample_viewstate}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.sample_jsf_notvuln}
        module_test.set_expect_requests(respond_args=respond_args)

        expect_args = {"uri": "/cookie.aspx"}
        respond_args = {
            "response_data": "<html><body><p>JWT Cookie Test</p></body></html>",
            "headers": {
                "set-cookie": "vulnjwt=eyJhbGciOiJIUzI1NiJ9.eyJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkJhZFNlY3JldHMiLCJleHAiOjE1OTMxMzM0ODMsImlhdCI6MTQ2NjkwMzA4M30.ovqRikAo_0kKJ0GVrAwQlezymxrLGjcEiW_s3UJMMCo; secure"
            },
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"uri": "/cookie2.aspx"}
        respond_args = {
            "response_data": "<html><body><p>Express Cookie Test (ES)</p></body></html>",
            "headers": {
                "set-cookie": "connect.sid=s%3A8FnPwdeM9kdGTZlWvdaVtQ0S1BCOhY5G.qys7H2oGSLLdRsEq7sqh7btOohHsaRKqyjV4LiVnBvc; Path=/; Expires=Wed, 05 Apr 2023 04:47:29 GMT; HttpOnly"
            },
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"uri": "/cookie3.aspx"}
        respond_args = {
            "response_data": "<html><body><p>Express Cookie Test (CS)</p></body></html>",
            "headers": {
                "set-cookie": [
                    "foo=eyJ1c2VybmFtZSI6IkJib3RJc0xpZmUifQ==; path=/; HttpOnly",
                    "foo.sig=zOQU7v7aTe_3zu7tnVuHi1MJ2DU; path=/; HttpOnly",
                ],
            },
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        SecretFound = False
        IdentifyOnly = False
        CookieBasedDetection = False
        CookieBasedDetection_2 = False
        CookieBasedDetection_3 = False

        for e in events:
            if (
                e.type == "VULNERABILITY"
                and "Known Secret Found." in e.data["description"]
                and "validationKey: 0F97BAE23F6F36801ABDB5F145124E00A6F795A97093D778EE5CD24F35B78B6FC4C0D0D4420657689C4F321F8596B59E83F02E296E970C4DEAD2DFE226294979 validationAlgo: SHA1 encryptionKey: 8CCFBC5B7589DD37DC3B4A885376D7480A69645DAEEC74F418B4877BEC008156 encryptionAlgo: AES"
                in e.data["description"]
            ):
                SecretFound = True

            if (
                e.type == "FINDING"
                and "AHo0wmLu5ceItIi+I7XkEi1GAb4h12WZ894pA+Z4OH7bco2jXEy1RSCWwjtJcZNbWPcvPqL5zzfl03DoeMZfGGX7a9PSv+fUT8MAeKNouAGj1dZuO8srXt8xZIGg+wPCWWCzcX6IhWOtgWUwiXeSojCDTKXklsYt+kAAAAk5wOsXvb2lTJoO0Q=="
                in e.data["description"]
            ):
                IdentifyOnly = True

            if (
                e.type == "VULNERABILITY"
                and "1234" in e.data["description"]
                and "eyJhbGciOiJIUzI1NiJ9.eyJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkJhZFNlY3JldHMiLCJleHAiOjE1OTMxMzM0ODMsImlhdCI6MTQ2NjkwMzA4M30.ovqRikAo_0kKJ0GVrAwQlezymxrLGjcEiW_s3UJMMCo"
                in e.data["description"]
            ):
                CookieBasedDetection = True

            if (
                e.type == "VULNERABILITY"
                and "keyboard cat" in e.data["description"]
                and "s%3A8FnPwdeM9kdGTZlWvdaVtQ0S1BCOhY5G.qys7H2oGSLLdRsEq7sqh7btOohHsaRKqyjV4LiVnBvc"
                in e.data["description"]
            ):
                CookieBasedDetection_2 = True

            if (
                e.type == "VULNERABILITY"
                and "Express.js Secret (cookie-session)" in e.data["description"]
                and "zOQU7v7aTe_3zu7tnVuHi1MJ2DU" in e.data["description"]
            ):
                CookieBasedDetection_3 = True

        assert SecretFound, "No secret found"
        assert IdentifyOnly, "No crypto product identified"
        assert CookieBasedDetection, "No JWT cookie vuln detected"
        assert CookieBasedDetection_2, "No Express.js cookie vuln detected"
        assert CookieBasedDetection_3, "No Express.js (cs dual cookies) vuln detected"


class TestBadSecrets_customsecrets(TestBadSecrets):
    config_overrides = {
        "modules": {
            "badsecrets": {
                "custom_secrets": tempwordlist(
                    [
                        "DEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF,DEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF"
                    ]
                )
            }
        }
    }

    sample_viewstate = """
    <form method="post" action="./query.aspx" id="form1">
<div class="aspNetHidden">
<input type="hidden" name="__VIEWSTATE" id="__VIEWSTATE" value="/wEPDwUJODExMDE5NzY5ZGS02CHaDxi5Kw19mPShbrrOUCJ4pA==" />
</div>

<div class="aspNetHidden">

    <input type="hidden" name="__VIEWSTATEGENERATOR" id="__VIEWSTATEGENERATOR" value="75BBA7D6" />
    <input type="hidden" name="__VIEWSTATEENCRYPTED" id="__VIEWSTATEENCRYPTED" value="" />
</div>
    </form>
</body>
</html>
"""

    def check(self, module_test, events):
        SecretFound = False
        for e in events:
            if (
                e.type == "VULNERABILITY"
                and "Known Secret Found." in e.data["description"]
                and "DEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF" in e.data["description"]
            ):
                SecretFound = True
        assert SecretFound, "No secret found"



from bbot.modules.base import BaseModule


class reflected_parameters(BaseModule):
    watched_events = ["WEB_PARAMETER"]
    produced_events = ["FINDING"]
    flags = ["active", "safe", "web-thorough"]
    meta = {
        "description": "Highlight parameters that reflect their contents in response body",
        "author": "@liquidsec",
        "created_date": "2024-10-29",
    }

    async def handle_event(self, event):
        url = event.data.get("url")
        reflection_detected = await self.detect_reflection(event, url)

        if reflection_detected:
            param_type = event.data.get("type", "UNKNOWN")
            description = (
                f"[{param_type}] Parameter value reflected in response body. Name: [{event.data['name']}] "
                f"Source Module: [{str(event.module)}]"
            )
            if event.data.get("original_value"):
                description += (
                    f" Original Value: [{self.helpers.truncate_string(str(event.data['original_value']), 200)}]"
                )
            data = {"host": str(event.host), "description": description, "url": url}
            await self.emit_event(data, "FINDING", event)

    async def detect_reflection(self, event, url):
        """Detects reflection by sending a probe with a random value and a canary parameter."""
        probe_parameter_name = event.data["name"]
        probe_parameter_value = self.helpers.rand_string()
        canary_parameter_value = self.helpers.rand_string()
        probe_response = await self.send_probe_with_canary(
            event,
            probe_parameter_name,
            probe_parameter_value,
            canary_parameter_value,
            cookies=event.data.get("assigned_cookies", {}),
            timeout=10,
        )

        # Check if the probe parameter value is reflected AND the canary is not
        if probe_response:
            response_text = probe_response.text
            reflection_result = probe_parameter_value in response_text and canary_parameter_value not in response_text
            return reflection_result
        return False

    async def send_probe_with_canary(self, event, parameter_name, parameter_value, canary_value, cookies, timeout=10):
        method = "GET"
        url = event.data["url"]
        headers = {}
        data = None
        json_data = None
        params = {parameter_name: parameter_value, "c4n4ry": canary_value}

        if event.data["type"] == "GETPARAM":
            url = f"{url}?{parameter_name}={parameter_value}&c4n4ry={canary_value}"
        elif event.data["type"] == "COOKIE":
            cookies.update(params)
        elif event.data["type"] == "HEADER":
            headers.update(params)
        elif event.data["type"] == "POSTPARAM":
            method = "POST"
            data = params
        elif event.data["type"] == "BODYJSON":
            method = "POST"
            json_data = params

        self.debug(
            f"Sending {method} request to {url} with headers: {headers}, cookies: {cookies}, data: {data}, json: {json_data}"
        )

        response = await self.helpers.request(
            method=method, url=url, headers=headers, cookies=cookies, data=data, json=json_data, timeout=timeout
        )
        return response



from .base import BaseLightfuzz

import regex as re


class xss(BaseLightfuzz):
    """
    Detects Reflected Cross-Site Scripting vulnerabilities across multiple contexts and techniques

    * Context Detection:
       - Between HTML Tags: <tag>injection</tag>
       - Within Tag Attributes: <tag attribute="injection">
       - Inside JavaScript: <script>var x = 'injection'</script>

    * Context-Specific Testing:
       - Between Tags: Tests basic HTML injection and tag creation
       - Tag Attributes: Tests quote escaping and JavaScript event handlers
       - JavaScript Context: Tests string delimiter breaking and script tag termination
       - Handles both single and double quote contexts in JavaScript

    Can often detect through WAFs, since it does not attempt to construct an exploitation payload
    """

    friendly_name = "Cross-Site Scripting"

    async def determine_context(self, cookies, html, random_string):
        """
        Determines the context of the random string in the HTML response.
        With XSS, the context is what kind part of the page the injection is occuring in, which determine what payloads might be successful

        https://portswigger.net/web-security/cross-site-scripting/contexts
        """
        between_tags = False
        in_tag_attribute = False
        in_javascript = False

        between_tags_regex = re.compile(
            rf"<(\/?\w+)[^>]*>.*?{random_string}.*?<\/?\w+>"
        )  # The between tags context is when the injection occurs between HTML tags
        in_tag_attribute_regex = re.compile(
            rf'<(\w+)\s+[^>]*?(\w+)="([^"]*?{random_string}[^"]*?)"[^>]*>'
        )  # The in tag attribute context is when the injection occurs in an attribute of an HTML tag
        in_javascript_regex = re.compile(
            rf"<script\b[^>]*>[^<]*(?:<(?!\/script>)[^<]*)*{random_string}[^<]*(?:<(?!\/script>)[^<]*)*<\/script>"
        )  # The in javascript context is when the injection occurs within a <script> tag

        between_tags_match = await self.lightfuzz.helpers.re.search(between_tags_regex, html)
        if between_tags_match:
            between_tags = True

        in_tag_attribute_match = await self.lightfuzz.helpers.re.search(in_tag_attribute_regex, html)
        if in_tag_attribute_match:
            in_tag_attribute = True

        in_javascript_match = await self.lightfuzz.helpers.re.search(in_javascript_regex, html)
        if in_javascript_match:
            in_javascript = True

        return between_tags, in_tag_attribute, in_javascript

    async def determine_javascript_quote_context(self, target, text):
        # Define and compile regex patterns for double and single quotes
        quote_patterns = {"double": re.compile(f'"[^"]*{target}[^"]*"'), "single": re.compile(f"'[^']*{target}[^']*'")}

        # Split the text by semicolons to isolate JavaScript statements
        statements = text.split(";")

        # This function checks if the target string is balanced within a JavaScript statement
        def is_balanced(section, target_index, quote_char):
            left = section[:target_index]
            right = section[target_index + len(target) :]
            return left.count(quote_char) % 2 == 0 and right.count(quote_char) % 2 == 0

        # For each javascript statement, attempt to determine the type of quote we are within, and therefore what will enable breaking out of it to result in a successful XSS
        for statement in statements:
            for quote_type, pattern in quote_patterns.items():
                match = await self.lightfuzz.helpers.re.search(pattern, statement)
                if match:
                    context = match.group(0)
                    target_index = context.find(target)
                    opposite_quote = "'" if quote_type == "double" else '"'
                    if is_balanced(context, target_index, opposite_quote):
                        return quote_type
        # If we have no matches, the target string is most likely not within quotes
        return "outside"

    async def check_probe(self, cookies, probe, match, context):
        # Send the defined probe and look for the expected match value in the response
        probe_result = await self.standard_probe(self.event.data["type"], cookies, probe)
        if probe_result and match in probe_result.text:
            self.results.append(
                {
                    "type": "FINDING",
                    "description": f"Possible Reflected XSS. Parameter: [{self.event.data['name']}] Context: [{context}] Parameter Type: [{self.event.data['type']}]",
                }
            )
            return True
        return False

    async def fuzz(self):
        lightfuzz_event = self.event.parent
        cookies = self.event.data.get("assigned_cookies", {})

        # If this came from paramminer_getparams and didn't have a http_reflection tag, we don't need to check again
        if (
            lightfuzz_event.type == "WEB_PARAMETER"
            and str(lightfuzz_event.module) == "paramminer_getparams"
            and "http-reflection" not in lightfuzz_event.tags
        ):
            self.debug("Got WEB_PARAMETER from paramminer, with no reflection tag - xss is not possible, aborting")
            return

        reflection = None
        random_string = self.lightfuzz.helpers.rand_string(8)

        reflection_probe_result = await self.standard_probe(self.event.data["type"], cookies, random_string)
        # before continuing, check if the random string is reflected in the response - a prerequisite for XSS
        if reflection_probe_result and random_string in reflection_probe_result.text:
            reflection = True

        if not reflection or reflection is False:
            return

        between_tags, in_tag_attribute, in_javascript = await self.determine_context(
            cookies, reflection_probe_result.text, random_string
        )
        self.debug(
            f"determine_context returned: between_tags [{between_tags}], in_tag_attribute [{in_tag_attribute}], in_javascript [{in_javascript}]"
        )
        tags = [
            "z",
            "svg",
            "img",
        ]  # These represent easy to exploit tags, along with an arbitrary tag which is less likely to be blocked
        if between_tags:
            for tag in tags:
                between_tags_probe = f"<{tag}>{random_string}</{tag}>"
                result = await self.check_probe(
                    cookies, between_tags_probe, between_tags_probe, f"Between Tags ({tag} tag)"
                )  # After reflection in the HTTP response, did the tags survive without url-encoding or other sanitization/escaping?
                if result is True:
                    break

        if in_tag_attribute:
            in_tag_attribute_probe = f'{random_string}"z'
            in_tag_attribute_match = f'{random_string}"z'
            await self.check_probe(
                cookies, in_tag_attribute_probe, in_tag_attribute_match, "Tag Attribute"
            )  # After reflection in the HTTP response, did the quote survive without url-encoding or other sanitization/escaping?

            in_tag_attribute_probe = f'{random_string}"z'
            in_tag_attribute_match = f'"{random_string}""z'
            await self.check_probe(
                cookies, in_tag_attribute_probe, in_tag_attribute_match, "Tag Attribute (autoquote)"
            )  # After reflection in the HTTP response, did the quote survive without url-encoding or other sanitization/escaping (and account for auto-quoting)

            in_tag_attribute_probe = f"javascript:{random_string}"
            in_tag_attribute_match = f'action="javascript:{random_string}'
            await self.check_probe(
                cookies, in_tag_attribute_probe, in_tag_attribute_match, "Form Action Injection"
            )  # After reflection in the HTTP response, did the javascript sch

        if in_javascript:
            in_javascript_probe = rf"</script><script>{random_string}</script>"
            result = await self.check_probe(
                cookies, in_javascript_probe, in_javascript_probe, "In Javascript"
            )  # After reflection in the HTTP response, did the script tags survive without url-encoding or other sanitization/escaping?
            if result is False:
                # To attempt this technique, we need to determine the type of quote we are within
                quote_context = await self.determine_javascript_quote_context(
                    random_string, reflection_probe_result.text
                )

                # Skip the test if the context is outside
                if quote_context == "outside":
                    return

                # Update probes based on the quote context
                if quote_context == "single":
                    in_javascript_escape_probe = rf"a\';zzzzz({random_string})\\"
                    in_javascript_escape_match = rf"a\\';zzzzz({random_string})\\"
                elif quote_context == "double":
                    in_javascript_escape_probe = rf"a\";zzzzz({random_string})\\"
                    in_javascript_escape_match = rf'a\\";zzzzz({random_string})\\'

                await self.check_probe(
                    cookies,
                    in_javascript_escape_probe,
                    in_javascript_escape_match,
                    f"In Javascript (escaping the escape character, {quote_context} quote)",
                )



from ipaddress import ip_network

from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_scan(
    events,
    helpers,
    monkeypatch,
    bbot_scanner,
):
    scan0 = bbot_scanner(
        "1.1.1.0",
        "1.1.1.1/31",
        "evilcorp.com",
        "test.evilcorp.com",
        blacklist=["1.1.1.1/28", "www.evilcorp.com"],
        modules=["ipneighbor"],
    )
    await scan0.load_modules()
    assert scan0.whitelisted("1.1.1.1")
    assert scan0.whitelisted("1.1.1.0")
    assert scan0.blacklisted("1.1.1.15")
    assert not scan0.blacklisted("1.1.1.16")
    assert scan0.blacklisted("1.1.1.1/30")
    assert not scan0.blacklisted("1.1.1.1/27")
    assert not scan0.in_scope("1.1.1.1")
    assert scan0.whitelisted("api.evilcorp.com")
    assert scan0.whitelisted("www.evilcorp.com")
    assert not scan0.blacklisted("api.evilcorp.com")
    assert scan0.blacklisted("asdf.www.evilcorp.com")
    assert scan0.in_scope("test.api.evilcorp.com")
    assert not scan0.in_scope("test.www.evilcorp.com")
    assert not scan0.in_scope("www.evilcorp.co.uk")
    j = scan0.json
    assert set(j["target"]["seeds"]) == {"1.1.1.0", "1.1.1.0/31", "evilcorp.com", "test.evilcorp.com"}
    # we preserve the original whitelist inputs
    assert set(j["target"]["whitelist"]) == {"1.1.1.0/32", "1.1.1.0/31", "evilcorp.com", "test.evilcorp.com"}
    # but in the background they are collapsed
    assert scan0.target.whitelist.hosts == {ip_network("1.1.1.0/31"), "evilcorp.com"}
    assert set(j["target"]["blacklist"]) == {"1.1.1.0/28", "www.evilcorp.com"}
    assert "ipneighbor" in j["preset"]["modules"]

    scan1 = bbot_scanner("1.1.1.1", whitelist=["1.0.0.1"])
    assert not scan1.blacklisted("1.1.1.1")
    assert not scan1.blacklisted("1.0.0.1")
    assert not scan1.whitelisted("1.1.1.1")
    assert scan1.whitelisted("1.0.0.1")
    assert scan1.in_scope("1.0.0.1")
    assert not scan1.in_scope("1.1.1.1")

    scan2 = bbot_scanner("1.1.1.1")
    assert not scan2.blacklisted("1.1.1.1")
    assert not scan2.blacklisted("1.0.0.1")
    assert scan2.whitelisted("1.1.1.1")
    assert not scan2.whitelisted("1.0.0.1")
    assert scan2.in_scope("1.1.1.1")
    assert not scan2.in_scope("1.0.0.1")

    dns_table = {
        "1.1.1.1.in-addr.arpa": {"PTR": ["one.one.one.one"]},
        "one.one.one.one": {"A": ["1.1.1.1"]},
    }

    # make sure DNS resolution works
    scan4 = bbot_scanner("1.1.1.1", config={"dns": {"minimal": False}})
    await scan4.helpers.dns._mock_dns(dns_table)
    events = []
    async for event in scan4.async_start():
        events.append(event)
    event_data = [e.data for e in events]
    assert "one.one.one.one" in event_data

    # make sure it doesn't work when you turn it off
    scan5 = bbot_scanner("1.1.1.1", config={"dns": {"minimal": True}})
    await scan5.helpers.dns._mock_dns(dns_table)
    events = []
    async for event in scan5.async_start():
        events.append(event)
    event_data = [e.data for e in events]
    assert "one.one.one.one" not in event_data

    for scan in (scan0, scan1, scan2, scan4, scan5):
        await scan._cleanup()

    scan6 = bbot_scanner("a.foobar.io", "b.foobar.io", "c.foobar.io", "foobar.io")
    assert len(scan6.dns_strings) == 1


@pytest.mark.asyncio
async def test_task_scan_handle_event_timeout(bbot_scanner):
    from bbot.modules.base import BaseModule

    # make a module that takes a long time to handle an event
    class LongModule(BaseModule):
        watched_events = ["IP_ADDRESS"]
        handled_event = False
        cancelled = False
        _name = "long"

        async def handle_event(self, event):
            self.handled_event = True
            try:
                await self.helpers.sleep(99999999)
            except asyncio.CancelledError:
                self.cancelled = True
                raise

    # same thing but handle_batch
    class LongBatchModule(BaseModule):
        watched_events = ["IP_ADDRESS"]
        handled_event = False
        canceled = False
        _name = "long_batch"
        _batch_size = 2

        async def handle_batch(self, *events):
            self.handled_event = True
            try:
                await self.helpers.sleep(99999999)
            except asyncio.CancelledError:
                self.cancelled = True
                raise

    # scan with both modules
    scan = bbot_scanner(
        "127.0.0.1",
        config={
            "module_handle_event_timeout": 5,
            "module_handle_batch_timeout": 5,
        },
    )
    await scan._prep()
    scan.modules["long"] = LongModule(scan=scan)
    scan.modules["long_batch"] = LongBatchModule(scan=scan)
    events = [e async for e in scan.async_start()]
    assert events
    assert any(e.data == "127.0.0.1" for e in events)
    # make sure both modules were called
    assert scan.modules["long"].handled_event
    assert scan.modules["long_batch"].handled_event
    # they should also be cancelled
    assert scan.modules["long"].cancelled
    assert scan.modules["long_batch"].cancelled


@pytest.mark.asyncio
async def test_url_extension_handling(bbot_scanner):
    scan = bbot_scanner(config={"url_extension_blacklist": ["css"], "url_extension_httpx_only": ["js"]})
    await scan._prep()
    assert scan.url_extension_blacklist == {"css"}
    assert scan.url_extension_httpx_only == {"js"}
    good_event = scan.make_event("https://evilcorp.com/a.txt", "URL", tags=["status-200"], parent=scan.root_event)
    bad_event = scan.make_event("https://evilcorp.com/a.css", "URL", tags=["status-200"], parent=scan.root_event)
    httpx_event = scan.make_event("https://evilcorp.com/a.js", "URL", tags=["status-200"], parent=scan.root_event)
    assert "blacklisted" not in bad_event.tags
    assert "httpx-only" not in httpx_event.tags
    result = await scan.ingress_module.handle_event(good_event)
    assert result is None
    result, reason = await scan.ingress_module.handle_event(bad_event)
    assert result is False
    assert reason == "event is blacklisted"
    assert "blacklisted" in bad_event.tags
    result = await scan.ingress_module.handle_event(httpx_event)
    assert result is None
    assert "httpx-only" in httpx_event.tags

    await scan._cleanup()


@pytest.mark.asyncio
async def test_speed_counter():
    from bbot.scanner.stats import SpeedCounter

    # counter with 1-second window
    counter = SpeedCounter(1)
    # 10 events spread across 2 seconds
    for i in range(10):
        counter.tick()
        await asyncio.sleep(0.2)
    # only 5 should show
    assert 4 <= counter.speed <= 5


@pytest.mark.asyncio
async def test_python_output_matches_json(bbot_scanner):
    import json

    scan = bbot_scanner(
        "blacklanternsecurity.com",
        config={"speculate": True, "dns": {"minimal": False}, "scope": {"report_distance": 10}},
    )
    await scan.helpers.dns._mock_dns({"blacklanternsecurity.com": {"A": ["127.0.0.1"]}})
    events = [e.json() async for e in scan.async_start()]
    output_json = scan.home / "output.json"
    json_events = []
    for line in open(output_json):
        json_events.append(json.loads(line))

    assert len(events) == 5
    scan_events = [e for e in events if e["type"] == "SCAN"]
    assert len(scan_events) == 2
    assert all(isinstance(e["data"]["status"], str) for e in scan_events)
    assert len([e for e in events if e["type"] == "DNS_NAME"]) == 1
    assert len([e for e in events if e["type"] == "ORG_STUB"]) == 1
    assert len([e for e in events if e["type"] == "IP_ADDRESS"]) == 1
    assert events == json_events


@pytest.mark.asyncio
async def test_huge_target_list(bbot_scanner, monkeypatch):
    # single target should only have one rule
    scan = bbot_scanner("evilcorp.com", config={"excavate": True})
    await scan._prep()
    assert "hostname_extraction_0" in scan.modules["excavate"].yara_rules_dict
    assert "hostname_extraction_1" not in scan.modules["excavate"].yara_rules_dict

    # over 10000 targets should be broken into two rules
    num_targets = 10005
    targets = [f"evil{i}.com" for i in range(num_targets)]
    scan = bbot_scanner(*targets, config={"excavate": True})
    await scan._prep()
    assert "hostname_extraction_0" in scan.modules["excavate"].yara_rules_dict
    assert "hostname_extraction_1" in scan.modules["excavate"].yara_rules_dict
    assert "hostname_extraction_2" not in scan.modules["excavate"].yara_rules_dict


@pytest.mark.asyncio
async def test_exclude_cdn(bbot_scanner, monkeypatch):
    # test that CDN exclusion works

    from bbot import Preset

    dns_mock = {
        "evilcorp.com": {"A": ["127.0.0.1"]},
        "www.evilcorp.com": {"A": ["127.0.0.1"]},
    }

    # first, run a scan with no CDN exclusion
    scan = bbot_scanner("evilcorp.com")
    await scan.helpers._mock_dns(dns_mock)

    from bbot.modules.base import BaseModule

    class DummyModule(BaseModule):
        watched_events = ["DNS_NAME"]

        async def handle_event(self, event):
            if event.type == "DNS_NAME" and event.data == "evilcorp.com":
                await self.emit_event("www.evilcorp.com", "DNS_NAME", parent=event, tags=["cdn-cloudflare"])
            if event.type == "DNS_NAME" and event.data == "www.evilcorp.com":
                await self.emit_event("www.evilcorp.com:80", "OPEN_TCP_PORT", parent=event, tags=["cdn-cloudflare"])
                await self.emit_event("www.evilcorp.com:443", "OPEN_TCP_PORT", parent=event, tags=["cdn-cloudflare"])
                await self.emit_event("www.evilcorp.com:8080", "OPEN_TCP_PORT", parent=event, tags=["cdn-cloudflare"])

    dummy = DummyModule(scan=scan)
    await scan._prep()
    scan.modules["dummy"] = dummy
    events = [e async for e in scan.async_start() if e.type in ("DNS_NAME", "OPEN_TCP_PORT")]
    assert set(e.data for e in events) == {
        "evilcorp.com",
        "www.evilcorp.com",
        "www.evilcorp.com:80",
        "www.evilcorp.com:443",
        "www.evilcorp.com:8080",
    }

    monkeypatch.setattr("sys.argv", ["bbot", "-t", "evilcorp.com", "--exclude-cdn"])

    # then run a scan with --exclude-cdn enabled
    preset = Preset("evilcorp.com")
    preset.parse_args()
    assert preset.bake().to_yaml() == "modules:\n- portfilter\n"
    scan = bbot_scanner("evilcorp.com", preset=preset)
    await scan.helpers._mock_dns(dns_mock)
    dummy = DummyModule(scan=scan)
    await scan._prep()
    scan.modules["dummy"] = dummy
    events = [e async for e in scan.async_start() if e.type in ("DNS_NAME", "OPEN_TCP_PORT")]
    assert set(e.data for e in events) == {
        "evilcorp.com",
        "www.evilcorp.com",
        "www.evilcorp.com:80",
        "www.evilcorp.com:443",
    }



from .base import ModuleTestBase

smuggler_text = r"""
              ______ _
     / _____)                       | |
    ( (____  ____  _   _  ____  ____| | _____  ____
     \____ \|    \| | | |/ _  |/ _  | || ___ |/ ___)
     _____) ) | | | |_| ( (_| ( (_| | || ____| |
    (______/|_|_|_|____/ \___ |\___ |\_)_____)_|
                        (_____(_____|

         @defparam v1.1

    [+] URL        : http://127.0.0.1:8888
    [+] Method     : POST
    [+] Endpoint   : /
    [+] Configfile : default.py
    [+] Timeout    : 5.0 seconds
    [+] Cookies    : 1 (Appending to the attack)
    [nameprefix1]  : Checking TECL...
    [nameprefix1]  : Checking CLTE...
    [nameprefix1]  : OK (TECL: 0.61 - 405) (CLTE: 0.62 - 405)
    [tabprefix1]   : Checking TECL...git
    [tabprefix1]   : Checking CLTE...
    [tabprefix1]   : Checking TECL...
    [tabprefix1]   : Checking CLTE...
    [tabprefix1]   : Checking TECL...
    [tabprefix1]   : Checking CLTE...
    [tabprefix1]   : Potential CLTE Issue Found - POST @ http://127.0.0.1:8888 - default.py
    [CRITICAL]     : CLTE Payload: /home/user/.bbot/tools/smuggler/payloads/http_127.0.0.1_net_CLTE_tabprefix1.txt URL: http://127.0.0.1:8888/
    """


class TestSmuggler(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "smuggler"]

    async def setup_after_prep(self, module_test):
        old_run_live = module_test.scan.helpers.run_live

        async def smuggler_mock_run_live(*command, **kwargs):
            if "smuggler" not in command[0][1]:
                async for l in old_run_live(*command, **kwargs):
                    yield l
            else:
                for line in smuggler_text.splitlines():
                    yield line

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", smuggler_mock_run_live)

        request_args = {"uri": "/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(request_args, respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and "[HTTP SMUGGLER] [Potential CLTE Issue Found] Technique:     [tabprefix1]" in e.data["description"]
            for e in events
        ), "Failed to parse mocked command output"



from .base import ModuleTestBase
import re


class TestAspnetBinExposure(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "aspnet_bin_exposure"]
    config_overrides = {
        "modules": {
            "aspnet_bin_exposure": {
                "test_dlls": [
                    "Newtonsoft.Json.dll",
                ]
            }
        }
    }

    async def setup_before_prep(self, module_test):
        # Simulate successful DLL exposure
        expect_args = {
            "method": "GET",
            "uri": "/b/(S(X))in/Newtonsoft.Json.dll/(S(X))/",
        }
        respond_args = {
            "status": 200,
            "headers": {"content-type": "application/x-msdownload"},
            "response_data": b"MZ\x90\x00\x03\x00\x00\x00",
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate failed DLL exposure (confirmation test)
        expect_args = {
            "method": "GET",
            "uri": "/b/(S(X))in/oopsnotarealdll.dll/(S(X))/",
        }
        respond_args = {"status": 404}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate alternative technique
        expect_args = {
            "method": "GET",
            "uri": "/(S(X))/b/(S(X))in/Newtonsoft.Json.dll",
        }
        respond_args = {
            "status": 200,
            "headers": {"content-type": "application/x-msdownload"},
            "response_data": b"MZ\x90\x00\x03\x00\x00\x00",
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate failed alternative technique (confirmation test)
        expect_args = {
            "method": "GET",
            "uri": "/(S(X))/b/(S(X))in/oopsnotarealdll.dll",
        }
        respond_args = {"status": 404}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Fallback for any other requests
        expect_args = {"uri": re.compile(r"^/.*$")}
        respond_args = {"status": 404}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        vulnerability_found = False
        for e in events:
            if e.type == "VULNERABILITY" and "IIS Bin Directory DLL Exposure" in e.data["description"]:
                vulnerability_found = True
                assert e.data["severity"] == "HIGH", "Vulnerability severity should be HIGH"
                assert "Detection Url" in e.data["description"], "Description should include detection URL"
                break

        assert vulnerability_found, "No vulnerability event was found"



from contextlib import suppress
from collections import defaultdict

from bbot.modules.output.base import BaseOutputModule


class Web_parameters(BaseOutputModule):
    watched_events = ["WEB_PARAMETER"]
    meta = {
        "description": "Output WEB_PARAMETER names to a file",
        "created_date": "2025-01-25",
        "author": "@liquidsec",
    }
    options = {"output_file": "", "include_count": False}
    options_desc = {
        "output_file": "Output to file",
        "include_count": "Include the count of each parameter in the output",
    }

    output_filename = "web_parameters.txt"

    async def setup(self):
        self._prep_output_dir(self.output_filename)
        self.parameter_counts = defaultdict(int)
        return True

    async def handle_event(self, event):
        parameter_name = event.data.get("name", "")
        if parameter_name:
            self.parameter_counts[parameter_name] += 1

    async def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    async def report(self):
        include_count = self.config.get("include_count", False)

        # Sort behavior:
        # - If include_count is True, sort by count (descending) and then alphabetically by name
        # - If include_count is False, sort alphabetically by name only
        sorted_parameters = sorted(
            self.parameter_counts.items(), key=lambda x: (-x[1], x[0]) if include_count else x[0]
        )
        for param, count in sorted_parameters:
            if include_count:
                # Include the count of each parameter in the output
                self.file.write(f"{count}\t{param}\n")
            else:
                # Only include the parameter name, effectively deduplicating by name
                self.file.write(f"{param}\n")
        self.file.flush()
        if getattr(self, "_file", None) is not None:
            self.info(f"Saved web parameters to {self.output_file}")
